import dash_ag_grid as dag
from dash import Dash, html, dcc
from . import utils
from dash.testing.wait import until


def test_cr001_class_rules(dash_duo):
    app = Dash(
        __name__,
        meta_tags=[
            {
                "http-equiv": "content-security-policy",
                "content": "default-src 'self'; script-src 'self' 'unsafe-inline';"
                " style-src https://* 'self' 'unsafe-inline'; "
                "font-src data: https://* 'self' 'unsafe-inline';"
                "img-src data: https://* 'self'",
            }
        ],
    )

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    columnDefs = [
        {
            "headerName": "Make",
            "field": "make",
            "cellClassRules": {
                "orange": "params.value == 'Ford'",
                "red": "params.value == 'Porsche'",
                "blue": "params.value == 'Toyota'",
            },
        },
        {"headerName": "Model", "field": "model"},
        {"headerName": "Price", "field": "price"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown(
                "Try filtering the data in the grid using the inline filters (click the hamburger menu in each column). Though the data is still in `rowData`, you can view the virtual row data in callbacks by watching the `virtualRowData` property."
            ),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef={
                    "editable": True,
                },
                rowClassRules={
                    "row-orange": "params.data.model == 'Mondeo'",
                    "row-red": "params.data.model == 'Boxster'",
                    "row-blue": "params.data.model == 'Celica'",
                },
                dashGridOptions={"getRowClass": {"function": "rowTest(params)"}},
            ),
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")

    ### testing class rules
    grid.get_cell(0, 0).click()
    until(lambda: "row-blue" in grid.get_row(0).get_attribute("class"), timeout=3)
    until(lambda: "testing" in grid.get_row(0).get_attribute("class"), timeout=3)
    until(lambda: "blue" in grid.get_cell(0, 0).get_attribute("class"), timeout=3)
    grid.get_cell(0, 0).send_keys("t")
    grid.get_cell(0, 1).send_keys("t")
    grid.get_cell(1, 0).click()
    until(lambda: "row-blue" not in grid.get_row(0).get_attribute("class"), timeout=3)
    until(lambda: "blue" not in grid.get_cell(0, 0).get_attribute("class"), timeout=3)

    until(lambda: "row-orange" in grid.get_row(1).get_attribute("class"), timeout=3)
    until(lambda: "orange" in grid.get_cell(1, 0).get_attribute("class"), timeout=3)
    grid.get_cell(1, 0).send_keys("t")
    grid.get_cell(1, 1).send_keys("t")
    grid.get_cell(2, 0).click()
    until(lambda: "row-orange" not in grid.get_row(1).get_attribute("class"), timeout=3)
    until(lambda: "orange" not in grid.get_cell(1, 0).get_attribute("class"), timeout=3)

    until(lambda: "row-red" in grid.get_row(2).get_attribute("class"), timeout=3)
    until(lambda: "red" in grid.get_cell(2, 0).get_attribute("class"), timeout=3)
    grid.get_cell(2, 0).send_keys("t")
    grid.get_cell(2, 1).send_keys("t")
    grid.get_cell(1, 0).click()
    until(lambda: "red" not in grid.get_row(2).get_attribute("class"), timeout=3)
    until(lambda: "red" not in grid.get_cell(2, 0).get_attribute("class"), timeout=3)



from dash import Dash, html, Output, Input, no_update, State, ctx
import dash_ag_grid as dag
import plotly.express as px
import json

from . import utils
from dash.testing.wait import until


df = px.data.election()
default_display_cols = ["district_id", "district", "winner"]


def test_us001_user_style(dash_duo):
    app = Dash(__name__)
    columnDefs = [
        {"headerName": "Make", "field": "make"},
        {"headerName": "Model", "field": "model"},
        {"headerName": "Price", "field": "price"},
    ]

    defaultColDef = {
        "initialWidth": 150,
        "sortable": True,
        "resizable": True,
        "filter": True,
    }

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnSize="autoSize",
                columnDefs=columnDefs,
                defaultColDef=defaultColDef,
                rowData=rowData,
                style={"height": "500px", "width": "500px"},
            ),
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")
    until(
        lambda: "height: 500px"
        in dash_duo.find_element("div.ag-theme-alpine").get_attribute("style"),
        timeout=3,
    )
    until(
        lambda: "width: 500px"
        in dash_duo.find_element("div.ag-theme-alpine").get_attribute("style"),
        timeout=3,
    )



import dash_ag_grid as dag
from dash import Dash, html, dcc
from . import utils
from dash.testing.wait import until


def test_cf001_conditional_formatting(dash_duo):
    app = Dash(__name__)

    columnDefs = [
        {
            "headerName": "Make",
            "field": "make",
        },
        {
            "headerName": "Model",
            "field": "model",
        },
        {"headerName": "Price", "field": "price"},
        {"field": "changes"},
    ]

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    cellStyle = {
        "styleConditions": [
            {"condition": "highlightEdits(params)", "style": {"color": "orange"}},
        ]
    }

    defaultColDef = {
        "valueSetter": {"function": "addEdits(params)"},
        "editable": True,
        "cellStyle": cellStyle,
    }

    getRowStyle = {
        "styleConditions": [
            {"condition": "params.data.make == 'Toyota'", "style": {"color": "blue"}}
        ]
    }

    app.layout = html.Div(
        [
            dcc.Markdown(
                "In this grid, the __Make__ column has a popup below the cell,  the __Model__ has a popup above the cell, and the __Price__ has the default (in cell) editor."
            ),
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef=defaultColDef,
                columnSize="sizeToFit",
                getRowStyle=getRowStyle,
                id="grid",
            ),
            html.Button(id="focus"),
        ],
        style={"margin": 20},
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")

    ### testing styles
    grid.get_cell(0, 0).click()
    until(lambda: "color: blue" in grid.get_row(0).get_attribute("style"), timeout=3)
    grid.get_cell(0, 0).send_keys("t")
    grid.get_cell(0, 1).click()
    until(
        lambda: "color: orange" in grid.get_cell(0, 0).get_attribute("style"), timeout=3
    )
    until(
        lambda: "color: blue" not in grid.get_row(0).get_attribute("style"), timeout=3
    )
    until(
        lambda: "color: orange" not in grid.get_cell(0, 1).get_attribute("style"),
        timeout=3,
    )
    grid.get_cell(0, 1).send_keys("t")
    grid.get_cell(0, 2).click()
    until(
        lambda: "color: orange" not in grid.get_cell(0, 2).get_attribute("style"),
        timeout=3,
    )
    until(
        lambda: "color: orange" in grid.get_cell(0, 1).get_attribute("style"), timeout=3
    )
    grid.get_cell(0, 2).send_keys("t")
    grid.get_cell(0, 0).click()
    until(
        lambda: "color: orange" in grid.get_cell(0, 2).get_attribute("style"), timeout=3
    )
    assert "color: orange" in grid.get_cell(0, 0).get_attribute("style")



import dash_ag_grid as dag
from dash import Dash, html, dcc, Output, Input, no_update, ctx, State
import requests
from . import utils
import json
from dash.testing.wait import until
import time


def test_sb001_sizing_buttons(dash_duo):
    app = Dash(__name__)

    data = requests.get(
        r"https://www.ag-grid.com/example-assets/olympic-winners.json"
    ).json()

    columnDefs = [
        {"field": "athlete"},
        {"field": "age"},
        {"field": "country"},
        {"field": "year"},
        {"field": "date"},
        {"field": "sport"},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
        {"field": "total"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown("This grid has single-select rows."),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=data[:100],
                columnSize="sizeToFit",
                defaultColDef={
                    "resizable": True,
                    "sortable": True,
                    "filter": True,
                    "floatingFilter": True,
                },
                dashGridOptions={"rowSelection": "multiple"},
                persistence=True,
                persistence_type="session",
            ),
            html.Button(id="autoSizeAllColumns", children="Auto Size All"),
            html.Button(
                id="autoSizeAllColumnsSkipHeaders", children="Auto Size All SkipHeaders"
            ),
            html.Div(id="columnState"),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Output("grid", "columnSizeOptions", allow_duplicate=True),
        Input("autoSizeAllColumns", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "autoSize", {"skipHeader": False}
        return no_update, no_update

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Output("grid", "columnSizeOptions", allow_duplicate=True),
        Input("autoSizeAllColumnsSkipHeaders", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "autoSize", {"skipHeader": True}
        return no_update, no_update

    @app.callback(
        Output("columnState", "children"),
        Input("grid", "columnState"),
        State("columnState", "children"),
        State("grid", "columnSizeOptions"),
    )
    def selected(state, oldState, opts):
        if state:
            test = True
            if oldState and opts == {"skipHeader": True}:
                oldState = json.loads(oldState)
                for i in range(len(state)):
                    if i in [1, 6, 7, 8, 9]:
                        if state[i]["width"] > oldState[i]["width"]:
                            test = False
                            break
            assert test
            return json.dumps(state)
        return ""

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    oldValue = ""
    until(lambda: oldValue != dash_duo.find_element("#columnState").text, timeout=3)
    oldValue = dash_duo.find_element("#columnState").text
    for x in columnDefs:
        assert x["field"] in oldValue

    for x in ["autoSizeAllColumns", "autoSizeAllColumnsSkipHeaders"]:
        dash_duo.find_element(f"#{x}").click()
        until(
            lambda: oldValue
            != dash_duo.find_element("#columnState").get_attribute("innerText"),
            timeout=3,
        )
        oldValue = dash_duo.find_element("#columnState").text


def test_sb002_sizing_buttons(dash_duo):
    app = Dash(__name__)

    data = requests.get(
        r"https://www.ag-grid.com/example-assets/olympic-winners.json"
    ).json()

    columnDefs = [
        {"field": "athlete"},
        {"field": "age"},
        {"field": "country"},
        {"field": "year"},
        {"field": "date"},
        {"field": "sport"},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
        {"field": "total"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown("This grid has single-select rows."),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=data[:100],
                columnSize="autoSize",
                defaultColDef={
                    "resizable": True,
                    "sortable": True,
                    "filter": True,
                    "floatingFilter": True,
                },
                dashGridOptions={"rowSelection": "multiple"},
                persistence=True,
                persistence_type="session",
            ),
            html.Button(id="sizeToFit", children="sizeToFit"),
            html.Button(id="responsiveSizeToFit", children="responsiveSizeToFit"),
            html.Div(id="columnState"),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Input("sizeToFit", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "sizeToFit"
        return no_update

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Input("responsiveSizeToFit", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "responsiveSizeToFit"
        return no_update

    @app.callback(
        Output("columnState", "children"),
        Input("grid", "columnState"),
    )
    def selected(state):
        if state:
            return json.dumps(state)
        return ""

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    dash_duo.driver.set_window_size(1000, 1000)
    oldValue = ""
    until(lambda: oldValue != dash_duo.find_element("#columnState").text, timeout=3)
    oldValue = dash_duo.find_element("#columnState").text
    for x in columnDefs:
        assert x["field"] in oldValue

    for x in ["sizeToFit", "responsiveSizeToFit"]:
        dash_duo.find_element(f"#{x}").click()
        if x != "responsiveSizeToFit":
            until(
                lambda: oldValue
                != dash_duo.find_element("#columnState").get_attribute("innerText"),
                timeout=3,
            )
        oldValue = dash_duo.find_element("#columnState").text
        dash_duo.driver.set_window_size(500, 500)
        if x == "responsiveSizeToFit":
            until(
                lambda: oldValue
                != dash_duo.find_element("#columnState").get_attribute("innerText"),
                timeout=3,
            )
        else:
            assert oldValue == dash_duo.find_element("#columnState").get_attribute(
                "innerText"
            )

        oldValue = dash_duo.find_element("#columnState").text
        dash_duo.driver.set_window_size(1000, 1000)


def test_sb003_sizing_buttons(dash_duo):
    app = Dash(__name__)

    data = requests.get(
        r"https://www.ag-grid.com/example-assets/olympic-winners.json"
    ).json()

    columnDefs = [
        {"field": "athlete"},
        {"field": "age"},
        {"field": "country"},
        {"field": "year"},
        {"field": "date"},
        {"field": "sport"},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
        {"field": "total"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown("This grid has single-select rows."),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=data[:100],
                columnSize="sizeToFit",
                defaultColDef={
                    "resizable": True,
                    "sortable": True,
                    "filter": True,
                    "floatingFilter": True,
                },
                dashGridOptions={"rowSelection": "multiple"},
                persistence=True,
                persistence_type="session",
            ),
            html.Button(id="autoSizeOneColumn", children="Auto Size One"),
            html.Button(
                id="autoSizeOneColumnSkipHeaders", children="Auto Size One SkipHeaders"
            ),
            html.Div(id="columnState"),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Output("grid", "columnSizeOptions", allow_duplicate=True),
        Input("autoSizeOneColumn", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "autoSize", {"keys": ["gold"]}
        return no_update, no_update

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Output("grid", "columnSizeOptions", allow_duplicate=True),
        Input("autoSizeOneColumnSkipHeaders", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "autoSize", {"skipHeader": True, "keys": ["gold"]}
        return no_update, no_update

    @app.callback(
        Output("columnState", "children"),
        Input("grid", "columnState"),
        State("columnState", "children"),
        State("grid", "columnSizeOptions"),
    )
    def selected(state, oldState, opts):
        if state:
            test = True
            if oldState and opts == {"skipHeader": True, "keys": ["gold"]}:
                oldState = json.loads(oldState)
                for i in range(len(state)):
                    if i in [6]:
                        if state[i]["width"] > oldState[i]["width"]:
                            test = False
                            break
            assert test
            return json.dumps(state)
        return ""

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    oldValue = ""
    until(lambda: oldValue != dash_duo.find_element("#columnState").text, timeout=3)
    oldValue = dash_duo.find_element("#columnState").text
    for x in columnDefs:
        assert x["field"] in oldValue

    for x in ["autoSizeOneColumn", "autoSizeOneColumnSkipHeaders"]:
        dash_duo.find_element(f"#{x}").click()
        until(
            lambda: oldValue
            != dash_duo.find_element("#columnState").get_attribute("innerText"),
            timeout=3,
        )
        oldValue = dash_duo.find_element("#columnState").text


def test_sb004_sizing_buttons(dash_duo):
    app = Dash(__name__)

    data = requests.get(
        r"https://www.ag-grid.com/example-assets/olympic-winners.json"
    ).json()

    columnDefs = [
        {"field": "athlete"},
        {"field": "age"},
        {"field": "country"},
        {"field": "year"},
        {"field": "date"},
        {"field": "sport"},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
        {"field": "total"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown("This grid has single-select rows."),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=data[:100],
                columnSize="autoSize",
                defaultColDef={
                    "resizable": True,
                    "sortable": True,
                    "filter": True,
                    "floatingFilter": True,
                },
                dashGridOptions={"rowSelection": "multiple"},
                persistence=True,
                persistence_type="session",
            ),
            html.Button(id="sizeToFit", children="sizeToFit"),
            html.Button(id="responsiveSizeToFit", children="responsiveSizeToFit"),
            html.Div(id="columnState"),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Output("grid", "columnSizeOptions", allow_duplicate=True),
        Input("sizeToFit", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "sizeToFit", {"defaultMaxWidth": 50}
        return no_update, no_update

    @app.callback(
        Output("grid", "columnSize", allow_duplicate=True),
        Output("grid", "columnSizeOptions", allow_duplicate=True),
        Input("responsiveSizeToFit", "n_clicks"),
        prevent_initial_call=True,
    )
    def setSelection(n):
        if n:
            return "responsiveSizeToFit", {"defaultMinWidth": 50}
        return no_update, no_update

    @app.callback(
        Output("columnState", "children"),
        Input("grid", "columnState"),
    )
    def selected(state):
        if state:
            return json.dumps(state)
        return ""

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    dash_duo.driver.set_window_size(1000, 1000)
    oldValue = ""
    until(lambda: oldValue != dash_duo.find_element("#columnState").text, timeout=3)
    oldValue = dash_duo.find_element("#columnState").text
    for x in columnDefs:
        assert x["field"] in oldValue

    for x in ["sizeToFit", "responsiveSizeToFit"]:
        dash_duo.find_element(f"#{x}").click()
        if x != "responsiveSizeToFit":
            until(
                lambda: oldValue
                != dash_duo.find_element("#columnState").get_attribute("innerText"),
                timeout=3,
            )
        oldValue = dash_duo.find_element("#columnState").text
        dash_duo.driver.set_window_size(400, 400)
        if x == "responsiveSizeToFit":
            until(
                lambda: oldValue
                != dash_duo.find_element("#columnState").get_attribute("innerText"),
                timeout=3,
            )
        else:
            assert oldValue == dash_duo.find_element("#columnState").get_attribute(
                "innerText"
            )
            dash_duo.find_element(f"#{x}").click()
            until(
                lambda: oldValue
                != dash_duo.find_element("#columnState").get_attribute("innerText"),
                timeout=3,
            )

        oldValue = dash_duo.find_element("#columnState").text
        dash_duo.driver.set_window_size(1000, 1000)



from dash import Dash, html, Input, Output
from dash_ag_grid import AgGrid
import plotly.express as px
import json

from . import utils


df = px.data.election()
default_display_cols = ["district_id", "district", "winner"]


def test_fi001_floating_filter(dash_duo):
    app = Dash()
    app.layout = html.Div(
        [
            AgGrid(
                id="grid",
                rowData=df.to_dict("records"),
                columnDefs=[
                    {"headerName": col.capitalize(), "field": col}
                    for col in default_display_cols
                ],
                defaultColDef={"filter": True, "floatingFilter": True},
                filterModel={
                    "district": {
                        "filterType": "text",
                        "type": "contains",
                        "filter": "12",
                    }
                },
            ),
            html.Div(id="filterModel"),
            html.Button(id="resetFilters", n_clicks=0),
        ]
    )

    @app.callback(
        Output("filterModel", "children"),
        Input("grid", "filterModel"),
    )
    def updateFilterModel(fM):
        return json.dumps(fM)

    @app.callback(
        Output("grid", "filterModel"),
        Input("resetFilters", "n_clicks"),
        prevent_initial_call=True,
    )
    def updateFilterModel(n):
        return {}

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 1, "112-DeLorimier")
    dash_duo.find_element("#resetFilters").click()

    grid.wait_for_cell_text(0, 1, "101-Bois-de-Liesse")
    dash_duo.wait_for_text_to_equal("#filterModel", "{}")

    grid.set_filter(0, "12")
    dash_duo.wait_for_text_to_equal(
        "#filterModel",
        '{"district_id": {"filterType": "number",'
        ' "type": "equals", "filter": 12}}',
    )

    grid.wait_for_cell_text(0, 1, "12-Saint-Sulpice")
    grid.wait_for_rendered_rows(1)

    dash_duo.find_element("#resetFilters").click()
    grid.wait_for_cell_text(0, 1, "101-Bois-de-Liesse")



"""
Nested tables.
"""

import dash_ag_grid as dag
import dash
from dash import Input, Output, html, dcc, Dash
from . import utils
from dash.testing.wait import until
import time


def test_rf001_recursive_functions(dash_duo):
    app = Dash(__name__)
    masterColumnDefs = [
        {
            "headerName": "Country",
            "field": "country",
            "cellRenderer": "agGroupCellRenderer",
        },
        {"headerName": "Region", "field": "region"},
        {"headerName": "Population", "field": "population"},
    ]

    detailColumnDefs = [
        {
            "headerName": "City",
            "field": "city",
            "valueGetter": {"function": "`**${1+2}**`"},
            "cellStyle": {"color": "red"},
            "cellRendererParams": {
                "innerRenderer": "markdown"
            }
        },
        {"headerName": "Pop. (City proper)", "field": "population_city"},
        {"headerName": "Pop. (Metro area)", "field": "population_metro"},
        {
            "headerName": "testFun",
            "children": [
                {
                    "headerName": "Pop. (Metro area)",
                    "field": "population_metro",
                    "valueGetter": {"function": "`**${1+2}**`"},
                    "cellRenderer": "markdown"
                },
                {
                    "headerName": "testing",
                    "children": [
                        {
                            "headerName": "So Much Fun",
                            "field": "population_metro",
                            "valueFormatter": {"function": "3+5"},
                        },
                    ],
                },
            ],
        },
    ]

    nextLevel = detailColumnDefs.copy()
    nextLevel[0]["cellRenderer"] = "agGroupCellRenderer"

    rowData = [
        {
            "country": "China",
            "region": "Asia",
            "population": 1411778724,
            "cities": [
                {
                    "city": "Shanghai",
                    "population_city": 24870895,
                    "population_metro": "NA",
                },
                {
                    "city": "Beijing",
                    "population_city": 21893095,
                    "population_metro": "NA",
                },
                {
                    "city": "Chongqing",
                    "population_city": 32054159,
                    "population_metro": "NA",
                },
            ],
        },
        {
            "country": "India",
            "region": "Asia",
            "population": 1383524897,
            "cities": [
                {
                    "city": "Delhi",
                    "population_city": 16753235,
                    "population_metro": 29000000,
                    "cities": [
                        {
                            "city": "Delhi",
                            "population_city": 16753235,
                            "population_metro": 29000000,
                        },
                        {
                            "city": "Mumbai",
                            "population_city": 12478447,
                            "population_metro": 24400000,
                        },
                        {
                            "city": "Kolkata",
                            "population_city": 4496694,
                            "population_metro": 14035959,
                        },
                    ],
                },
                {
                    "city": "Mumbai",
                    "population_city": 12478447,
                    "population_metro": 24400000,
                },
                {
                    "city": "Kolkata",
                    "population_city": 4496694,
                    "population_metro": 14035959,
                },
            ],
        },
        {
            "country": "United States",
            "region": "Americas",
            "population": 332593407,
            "cities": [
                {
                    "city": "New York",
                    "population_city": 8398748,
                    "population_metro": 19303808,
                },
                {
                    "city": "Los Angeles",
                    "population_city": 3990456,
                    "population_metro": 13291486,
                },
                {
                    "city": "Chicago",
                    "population_city": 2746388,
                    "population_metro": 9618502,
                },
            ],
        },
        {
            "country": "Indonesia",
            "region": "Asia",
            "population": 271350000,
            "cities": [
                {
                    "city": "Jakarta",
                    "population_city": 10154134,
                    "population_metro": 33430285,
                },
            ],
        },
    ]

    cellStyle = {
        "styleConditions": [
            {"condition": 'params.data.city=="Delhi"', "style": {"color": "orange"}}
        ]
    }

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=masterColumnDefs,
                defaultColDef={"cellStyle": cellStyle},
                rowData=rowData,
                columnSize="sizeToFit",
                enableEnterpriseModules=True,
                masterDetail=True,
                detailCellRendererParams={
                    "detailGridOptions": {
                        "columnDefs": nextLevel,
                        "defaultColDef": {"cellStyle": cellStyle},
                        "detailCellRendererParams": {
                            "detailGridOptions": {
                                "columnDefs": detailColumnDefs,
                                "defaultColDef": {"cellStyle": cellStyle},
                            },
                            "detailColName": "cities",
                            "suppressCallback": True,
                        },
                        "masterDetail": True,
                    },
                    "detailColName": "cities",
                    "suppressCallback": True,
                },
                dashGridOptions={"detailRowAutoHeight": True},
            )
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "China")

    grid.get_cell_expandable(1, 0).click()

    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"]', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"] strong', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="4"]', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="4"] strong', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="5"]', "8"
    )
    assert "color: red" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="2"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="4"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="5"]'
    ).get_attribute("style")

    dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"] .ag-group-contracted'
    ).click()
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="1"]',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="1"] strong',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="4"]',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="4"] strong',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="5"]',
        "8",
    )
    assert "color: red" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="1"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="2"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="4"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="5"]'
    ).get_attribute("style")


def test_rf002_recursive_functions_server(dash_duo):
    app = Dash(__name__)
    masterColumnDefs = [
        {
            "headerName": "Country",
            "field": "country",
            "cellRenderer": "agGroupCellRenderer",
        },
        {"headerName": "Region", "field": "region"},
        {"headerName": "Population", "field": "population"},
    ]

    detailColumnDefs = [
        {
            "headerName": "City",
            "field": "city",
            "valueGetter": {"function": "`**${1+2}**`"},
            "cellRendererParams": {"innerRenderer": "markdown"},
            "cellStyle": {"color": "red"},
        },
        {"headerName": "Pop. (City proper)", "field": "population_city"},
        {"headerName": "Pop. (Metro area)", "field": "population_metro"},
        {
            "headerName": "testFun",
            "children": [
                {
                    "headerName": "Pop. (Metro area)",
                    "field": "population_metro",
                    "valueGetter": {"function": "`**${1+2}**`"},
                    "cellRenderer": "markdown"
                },
                {
                    "headerName": "testing",
                    "children": [
                        {
                            "headerName": "So Much Fun",
                            "field": "population_metro",
                            "valueFormatter": {"function": "3+5"},
                        },
                    ],
                },
            ],
        },
    ]

    nextLevel = detailColumnDefs.copy()
    nextLevel[0]["cellRenderer"] = "agGroupCellRenderer"

    rowData = [
        {
            "country": "China",
            "region": "Asia",
            "population": 1411778724,
            "cities": [
                {
                    "city": "Shanghai",
                    "population_city": 24870895,
                    "population_metro": "NA",
                },
                {
                    "city": "Beijing",
                    "population_city": 21893095,
                    "population_metro": "NA",
                },
                {
                    "city": "Chongqing",
                    "population_city": 32054159,
                    "population_metro": "NA",
                },
            ],
        },
        {
            "country": "India",
            "region": "Asia",
            "population": 1383524897,
            "cities": [
                {
                    "city": "Delhi",
                    "population_city": 16753235,
                    "population_metro": 29000000,
                    "cities": [
                        {
                            "city": "Delhi",
                            "population_city": 16753235,
                            "population_metro": 29000000,
                        },
                        {
                            "city": "Mumbai",
                            "population_city": 12478447,
                            "population_metro": 24400000,
                        },
                        {
                            "city": "Kolkata",
                            "population_city": 4496694,
                            "population_metro": 14035959,
                        },
                    ],
                },
                {
                    "city": "Mumbai",
                    "population_city": 12478447,
                    "population_metro": 24400000,
                },
                {
                    "city": "Kolkata",
                    "population_city": 4496694,
                    "population_metro": 14035959,
                },
            ],
        },
        {
            "country": "United States",
            "region": "Americas",
            "population": 332593407,
            "cities": [
                {
                    "city": "New York",
                    "population_city": 8398748,
                    "population_metro": 19303808,
                },
                {
                    "city": "Los Angeles",
                    "population_city": 3990456,
                    "population_metro": 13291486,
                },
                {
                    "city": "Chicago",
                    "population_city": 2746388,
                    "population_metro": 9618502,
                },
            ],
        },
        {
            "country": "Indonesia",
            "region": "Asia",
            "population": 271350000,
            "cities": [
                {
                    "city": "Jakarta",
                    "population_city": 10154134,
                    "population_metro": 33430285,
                },
            ],
        },
    ]

    cellStyle = {
        "styleConditions": [
            {"condition": 'params.data.city=="Delhi"', "style": {"color": "orange"}}
        ]
    }

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=masterColumnDefs,
                defaultColDef={"cellStyle": cellStyle},
                rowData=rowData,
                columnSize="sizeToFit",
                enableEnterpriseModules=True,
                masterDetail=True,
                detailCellRendererParams={
                    "detailGridOptions": {
                        "columnDefs": nextLevel,
                        "defaultColDef": {"cellStyle": cellStyle},
                        "detailCellRendererParams": {
                            "detailGridOptions": {
                                "columnDefs": detailColumnDefs,
                                "defaultColDef": {"cellStyle": cellStyle},
                            },
                            "suppressCallback": False,
                        },
                        "masterDetail": True,
                    },
                    "suppressCallback": False,
                },
                dashGridOptions={"detailRowAutoHeight": True},
            )
        ]
    )

    @app.callback(
        Output("grid", "getDetailResponse"),
        Input("grid", "getDetailRequest"),
        prevent_initial_call=True,
    )
    def handle_request(request):
        return request["data"]["cities"]

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "China")

    grid.get_cell_expandable(1, 0).click()

    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"]', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"] strong', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="4"]', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="4"] strong', "3"
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="5"]', "8"
    )
    assert "color: red" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="2"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="4"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="5"]'
    ).get_attribute("style")

    dash_duo.find_element(
        '#grid .ag-details-grid [row-index="0"] [aria-colindex="1"] .ag-group-contracted'
    ).click()
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="1"]',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="1"] strong',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="4"]',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="4"] strong',
        "3",
    )
    dash_duo.wait_for_text_to_equal(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="5"]',
        "8",
    )
    assert "color: red" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="1"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="2"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="4"]'
    ).get_attribute("style")
    assert "color: orange" in dash_duo.find_element(
        '#grid .ag-details-grid .ag-details-grid [row-index="0"] [aria-colindex="5"]'
    ).get_attribute("style")



from dash import Dash, html, dcc, Output, Input
import dash_ag_grid as dag

from . import utils


def test_dt001_dangerous_templates(dash_duo):
    columnDefs = [
        {
            "headerName": "Stock Ticker",
            "field": "ticker",
        },
        {
            "headerName": "Company",
            "field": "company",
        },
        {
            "headerComponentParams": {"template": "<div>Testing</div>"},
            "field": "price",
            "valueFormatter": {"function": """d3.format("($,.2f")(params.value)"""},
        },
    ]

    grid = dag.AgGrid(
        id="grid",
        columnDefs=columnDefs,
        columnSize="sizeToFit",
        dashGridOptions={
            "overlayLoadingTemplate": "<div>CustomLoadingOverlay</div>",
            "overlayNoRowsTemplate": "<div>whoops no info</div>",
        },
    )

    app = Dash(__name__)

    app.layout = html.Div(
        [
            dcc.Markdown("Example of custom loading overlay"),
            grid,
            html.Button("load", id="loadBlank", n_clicks=0),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("grid", "rowData"),
        Input("loadBlank", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadBlank(n):
        if n:
            return []

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    assert grid.get_header_cell(0).text == "Stock Ticker"

    dash_duo.wait_for_text_to_equal(".ag-overlay-loading-center", "Loading...")
    dash_duo.find_element("#loadBlank").click()
    dash_duo.wait_for_text_to_equal(".ag-overlay-no-rows-wrapper", "No Rows To Show")
    assert grid.get_header_cell(2).text == "Price"


def test_dt002_dangerous_templates(dash_duo):
    columnDefs = [
        {
            "headerName": "Stock Ticker",
            "field": "ticker",
        },
        {
            "headerName": "Company",
            "field": "company",
        },
        {
            "headerComponentParams": {"template": "<div>Testing</div>"},
            "field": "price",
            "valueFormatter": {"function": """d3.format("($,.2f")(params.value)"""},
        },
    ]

    grid = dag.AgGrid(
        id="grid",
        columnDefs=columnDefs,
        columnSize="sizeToFit",
        dashGridOptions={
            "overlayLoadingTemplate": "<div>CustomLoadingOverlay</div>",
            "overlayNoRowsTemplate": "<div>whoops no info</div>",
        },
        dangerously_allow_code=True,
    )

    app = Dash(__name__)

    app.layout = html.Div(
        [
            dcc.Markdown("Example of custom loading overlay"),
            grid,
            html.Button("load", id="loadBlank", n_clicks=0),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("grid", "rowData"),
        Input("loadBlank", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadBlank(n):
        if n:
            return []

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    assert grid.get_header_cell(0).text == "Stock Ticker"

    dash_duo.wait_for_text_to_equal(
        ".ag-overlay-loading-wrapper", "CustomLoadingOverlay"
    )
    dash_duo.find_element("#loadBlank").click()
    dash_duo.wait_for_text_to_equal(".ag-overlay-no-rows-wrapper", "whoops no info")
    assert grid.get_header_cell(2).text == "Testing"



import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output
from . import utils
from dash.testing.wait import until
import json


def test_cc001_cell_clicked(dash_duo):
    app = Dash(__name__)

    columnDefs = [
        {
            "headerName": "Make",
            "field": "make",
        },
        {
            "headerName": "Model",
            "field": "model",
        },
        {"headerName": "Price", "field": "price"},
    ]

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    defaultColDef = {"valueSetter": {"function": "addEdits(params)"}, "editable": True}

    app.layout = html.Div(
        [
            dcc.Markdown(
                "In this grid, the __Make__ column has a popup below the cell,  the __Model__ has a popup above the cell, and the __Price__ has the default (in cell) editor."
            ),
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef=defaultColDef,
                columnSize="sizeToFit",
                getRowId="params.data.model",
                id="grid",
            ),
            html.Div(id="focus"),
        ],
        style={"margin": 20},
    )

    @app.callback(Output("focus", "children"), Input("grid", "cellClicked"))
    def clickData(d):
        return json.dumps(d)

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")

    ### testing styles
    grid.get_cell(0, 0).click()
    until(
        lambda: '{"value": "Toyota", "colId": "make", "rowIndex": 0, "rowId": "Celica"'
        in dash_duo.find_element("#focus").get_attribute("innerText"),
        timeout=3,
    )
    grid.get_cell(1, 1).click()
    until(
        lambda: '{"value": "Mondeo", "colId": "model", "rowIndex": 1, "rowId": "Mondeo"'
        in dash_duo.find_element("#focus").get_attribute("innerText"),
        timeout=3,
    )
    grid.get_cell(2, 2).click()
    until(
        lambda: '{"value": 72000, "colId": "price", "rowIndex": 2, "rowId": "Boxster"'
        in dash_duo.find_element("#focus").get_attribute("innerText"),
        timeout=3,
    )



import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output
import json
from . import utils
from dash.testing.wait import until


def test_ga001_grid_apis(dash_duo):
    app = Dash(
        __name__,
        meta_tags=[
            {
                "http-equiv": "content-security-policy",
                "content": "default-src 'self'; script-src 'self' 'unsafe-inline';"
                           " style-src https://* 'self' 'unsafe-inline'; "
                           "font-src data: https://* 'self' 'unsafe-inline';"
                           "img-src data: https://* 'self'",
            }
        ],
    )

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    columnDefs = [
        {
            "headerName": "Make",
            "field": "make",
            "cellClassRules": {
                "orange": "params.value == 'Ford'",
                "red": "params.value == 'Porsche'",
                "blue": "params.value == 'Toyota'",
            },
        },
        {"headerName": "Model", "field": "model"},
        {"headerName": "Price", "field": "price"},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=[rowData[0]],
                defaultColDef={
                    "editable": True,
                },
                rowClassRules={
                    "row-orange": "params.data.model == 'Mondeo'",
                    "row-red": "params.data.model == 'Boxster'",
                    "row-blue": "params.data.model == 'Celica'",
                },
                dashGridOptions={"getRowClass": {"function": "rowTest(params)"},
                                 "preventDefaultOnContextMenu": True},
            ),
            html.Button(id='addRow', children='add_row', n_clicks=0),
            html.Button(id='hidePrice', n_clicks=0),
            html.Div(id='virtualRowData'),
            html.Div(id='columnState'),
            html.Div(id='cellContext')
        ]
    )

    app.clientside_callback("function (n) { if (n) {" \
                            "grid = dash_ag_grid.getApi('grid'); " \
                            "if (n == 1) {" \
                            "grid.applyTransactionAsync({'add': [" + json.dumps(rowData[1]) + "]});" \
                                                                                              "} else {" \
                                                                                              "grid.applyTransactionAsync({'add': [" + json.dumps(
        rowData[2]) + "]});" \
                      "}} return dash_clientside.no_update}", Output('addRow', 'id'), Input('addRow', 'n_clicks'),
                            prevent_intial_Call=True)

    app.clientside_callback(
        """function (d) {
            return JSON.stringify(d)
        }""", Output('virtualRowData', 'children'), Input('grid', 'virtualRowData')
    )

    app.clientside_callback(
        """function (d) {
            return JSON.stringify(d)
        }""", Output('columnState', 'children'), Input('grid', 'columnState')
    )

    app.clientside_callback(
        """async function (id) {
            grid = await dash_ag_grid.getApiAsync(id)
            grid.addEventListener('cellContextMenu', (params) => {
                    document.getElementById('cellContext').innerText = params.value;
                })
            return window.dash_clientside.no_update
        }""",
        Output('grid', 'id'), Input('grid', 'id')
    )

    app.clientside_callback(
        """function (n) {
            if (n) {
                dash_ag_grid.getApi('grid').setColumnVisible("price", false)
            }
            return dash_clientside.no_update
        }""",
        Output('hidePrice', 'id'), Input('hidePrice', 'n_clicks')
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")

    # testing adding data
    dash_duo.find_element("#addRow").click()
    grid.wait_for_cell_text(1, 0, "Ford")

    dash_duo.find_element("#addRow").click()
    grid.wait_for_cell_text(2, 0, "Porsche")

    until(lambda: json.loads(dash_duo.find_element('#virtualRowData').text) == rowData, timeout=3)

    dash_duo.find_element("#hidePrice").click()
    until(lambda: dash_duo.find_element('#columnState').text != '', timeout=3)
    until(lambda: json.loads(dash_duo.find_element('#columnState').text)[2]['hide'], timeout=3)

    action = utils.ActionChains(dash_duo.driver)
    action.context_click(grid.get_cell(0,0)).perform()
    until(lambda: dash_duo.find_element('#cellContext').text == 'Toyota', timeout=3)



from dash import Dash, html, Output, Input, no_update, State
from dash_ag_grid import AgGrid
import plotly.express as px
import pandas as pd

from . import utils


df = px.data.election()
default_display_cols = ["district_id", "district", "winner"]


def test_cd001_drag_columns(dash_duo):
    app = Dash()
    app.layout = html.Div(
        [
            AgGrid(
                id="grid",
                rowData=df.to_dict("records"),
                columnDefs=[
                    {"headerName": col.capitalize(), "field": col}
                    for col in default_display_cols
                ],
            )
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_all_header_texts(["District_id", "District", "Winner"])
    grid.wait_for_pinned_cols(0)
    grid.wait_for_viewport_cols(3)

    grid.drag_col(2, 0)  # last column first but not pinned

    grid.wait_for_all_header_texts(["Winner", "District_id", "District"])
    grid.wait_for_pinned_cols(0)
    grid.wait_for_viewport_cols(3)

    grid.pin_col(1)  # middle column pinned

    grid.wait_for_all_header_texts(["District_id", "Winner", "District"])
    grid.wait_for_pinned_cols(1)
    grid.wait_for_viewport_cols(2)

    # pin first non-pinned column by dragging it to its own left edge
    grid.pin_col(1, 1)

    grid.wait_for_all_header_texts(["District_id", "Winner", "District"])
    grid.wait_for_pinned_cols(2)
    grid.wait_for_viewport_cols(1)


def test_cd002_column_drag(dash_duo):
    data = {
        "ticker": ["AAPL", "MSFT", "AMZN", "GOOGL"],
        "company": ["Apple", "Microsoft", "Amazon", "Alphabet"],
        "price": [154.99, 268.65, 100.47, 96.75],
        "buy": ["Buy" for _ in range(4)],
        "sell": ["Sell" for _ in range(4)],
        "watch": ["Watch" for _ in range(4)],
    }
    df = pd.DataFrame(data)

    columnDefs = [
        {
            "headerName": "Stock Ticker",
            "field": "ticker",
        },
        {"headerName": "Company", "field": "company", "filter": True},
        {
            "headerName": "Last Close Price",
            "type": "rightAligned",
            "field": "price",
            "valueFormatter": {"function": """d3.format("($,.2f")(params.value)"""},
        },
    ]

    defaultColDef = {
        "resizable": True,
        "sortable": True,
        "editable": False,
    }

    grid = AgGrid(
        id="topGrid",
        columnDefs=columnDefs,
        rowData=df.to_dict("records"),
        columnSize="autoSize",
        defaultColDef=defaultColDef,
        dashGridOptions={"alignedGrids": ["middleGrid", "bottomGrid"]},
    )

    gridBot = AgGrid(
        id="bottomGrid",
        columnDefs=columnDefs,
        rowData=df.to_dict("records"),
        columnSize="autoSize",
        defaultColDef=defaultColDef,
    )

    app = Dash(__name__)

    app.layout = html.Div(
        [grid, html.Div(id="hiding"), gridBot, html.Button(id="link")],
        style={"margin": 20},
    )

    @app.callback(
        Output("hiding", "children"),
        Input("link", "n_clicks"),
        State("topGrid", "columnState"),
    )
    def addGrid(n, s):
        grid2 = AgGrid(
            id="middleGrid",
            columnDefs=columnDefs,
            rowData=df.to_dict("records"),
            columnSize=None,
            defaultColDef=defaultColDef,
            dashGridOptions={"alignedGrids": "bottomGrid"},
            columnState=s,
        )
        if n:
            return grid2
        return no_update

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "topGrid")
    botGrid = utils.Grid(dash_duo, "bottomGrid")

    grid.wait_for_all_header_texts(["Stock Ticker", "Company", "Last Close Price"])
    botGrid.wait_for_all_header_texts(["Stock Ticker", "Company", "Last Close Price"])
    grid.wait_for_pinned_cols(0)
    grid.wait_for_viewport_cols(3)
    botGrid.wait_for_pinned_cols(0)
    botGrid.wait_for_viewport_cols(3)

    grid.drag_col(2, 0)  # last column first but not pinned

    grid.wait_for_all_header_texts(["Last Close Price", "Stock Ticker", "Company"])
    botGrid.wait_for_all_header_texts(["Last Close Price", "Stock Ticker", "Company"])
    grid.wait_for_pinned_cols(0)
    grid.wait_for_viewport_cols(3)
    botGrid.wait_for_pinned_cols(0)
    botGrid.wait_for_viewport_cols(3)

    grid.pin_col(1)  # middle column pinned

    grid.wait_for_all_header_texts(["Stock Ticker", "Last Close Price", "Company"])
    botGrid.wait_for_all_header_texts(["Stock Ticker", "Last Close Price", "Company"])
    grid.wait_for_pinned_cols(1)
    botGrid.wait_for_pinned_cols(1)
    grid.wait_for_viewport_cols(2)
    botGrid.wait_for_viewport_cols(2)

    # add midGrid
    dash_duo.find_element("#link").click()

    # pin first non-pinned column by dragging it to its own left edge
    grid.pin_col(1, 1)
    midGrid = utils.Grid(dash_duo, "middleGrid")
    grid.wait_for_all_header_texts(["Stock Ticker", "Last Close Price", "Company"])
    midGrid.wait_for_all_header_texts(["Stock Ticker", "Last Close Price", "Company"])
    botGrid.wait_for_all_header_texts(["Stock Ticker", "Last Close Price", "Company"])
    grid.wait_for_pinned_cols(2)
    grid.wait_for_viewport_cols(1)
    botGrid.wait_for_pinned_cols(2)
    botGrid.wait_for_viewport_cols(1)
    midGrid.wait_for_pinned_cols(2)
    midGrid.wait_for_viewport_cols(1)

    # pin first non-pinned column by dragging it to its own left edge
    midGrid.pin_col(2, 2)

    grid.wait_for_all_header_texts(["Stock Ticker", "Last Close Price", "Company"])
    botGrid.wait_for_all_header_texts(["Stock Ticker", "Last Close Price", "Company"])
    grid.wait_for_pinned_cols(2)
    grid.wait_for_viewport_cols(1)
    botGrid.wait_for_pinned_cols(3)
    botGrid.wait_for_viewport_cols(0)
    midGrid.wait_for_pinned_cols(3)
    midGrid.wait_for_viewport_cols(0)



from selenium.webdriver import Keys

import dash_ag_grid as dag
from dash import Dash, html
from . import utils


def test_cd001_cell_data_types_override(dash_duo):
    app = Dash(__name__)

    rowData = [
        {"weight": 0.074657, "date": "01/01/2024"},
        {"weight": 0.06948567, "date": "02/01/2024"},
        {"weight": 0.02730574, "date": "03/01/2024"},
        {"weight": 0.0182345, "date": "04/01/2024"},
    ]

    columnDefs = [
        {"field": "weight", "cellDataType": "percentage"},
        {"field": "date", "cellDataType": "dateString"},
    ]

    # Only for second grid
    dataTypeDefinitions = {
        "percentage": {
            "baseDataType": "number",
            "extendsDataType": "number",
            "valueFormatter": {
                "function": "params.value == null ? '' :  d3.format(',.1%')(params.value)"
            },
        },
        "dateString": {
            "baseDataType": 'dateString',
            "extendsDataType": 'dateString',
            "valueParser": {
                "function": r"params.newValue != null && !!params.newValue.match(/\d{2}\/\d{2}\/\d{4}/) ? params.newValue : null"
            },
            "valueFormatter": {"function": "params.value == null ? '' : params.value"},
            "dataTypeMatcher": {"function": r"params != null && !!params.match(/\d{2}\/\d{2}\/\d{4}/)"},
            "dateParser": {"function": "dateParser(params)"},
            "dateFormatter": {"function": "dateFormatter(params)"},
        },
    }
    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid-cell-data-types-override-full-JS",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef={"filter": True, "editable": True},
                dashGridOptions={"dataTypeDefinitions": {"function": "dataTypeDefinitions"}},
            ),
            dag.AgGrid(
                id="grid-cell-data-types-override",
                columnDefs=columnDefs,
                rowData=rowData,
                defaultColDef={"filter": True, "editable": True},
                dashGridOptions={"dataTypeDefinitions": dataTypeDefinitions},
            ),
        ],
    )

    dash_duo.start_server(app)

    action = utils.ActionChains(dash_duo.driver)

    # same tests for both grids
    for id in ["grid-cell-data-types-override-full-JS", "grid-cell-data-types-override"]:
        grid = utils.Grid(dash_duo, id)

        # test overriden number cell data type
        action.double_click(grid.get_cell(0, 0)).perform()
        date_input_element = dash_duo.find_element(f'#{grid.id} .ag-number-field-input')
        date_input_element.send_keys("0.1" + Keys.ENTER)

        grid.wait_for_cell_text(0, 0, "10.0%")

        # test overriden dateString cell data type
        action.double_click(grid.get_cell(0, 1)).perform()
        date_input_element = dash_duo.find_element(f'#{grid.id} .ag-date-field-input')
        date_input_element.send_keys("01172024" + Keys.ENTER)

        grid.wait_for_cell_text(0, 1, "17/01/2024")



from dash import Dash, html, Output, Input, no_update, State, ctx
import dash_ag_grid as dag
import plotly.express as px
import json
import time

from . import utils
from dash.testing.wait import until


df = px.data.election()
default_display_cols = ["district_id", "district", "winner"]


def test_cs001_column_state(dash_duo):
    app = Dash(__name__)
    columnDefs = [
        {"headerName": "Make", "field": "make"},
        {"headerName": "Model", "field": "model"},
        {"headerName": "Price", "field": "price"},
    ]

    alt_columnDefs = [
        {"field": "price", "pinned": False, "sort": "asc"},
        {"field": "model", "pinned": False},
        {"field": "make", "pinned": False},
    ]

    defaultColDef = {
        "initialWidth": 150,
        "sortable": True,
        "resizable": True,
        "filter": True,
    }

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    colState = [
        {
            "colId": "make",
            "width": 150,
            "hide": False,
            "pinned": "left",
            "sort": None,
            "sortIndex": None,
            "aggFunc": None,
            "rowGroup": False,
            "rowGroupIndex": None,
            "pivot": False,
            "pivotIndex": None,
            "flex": None,
        },
        {
            "colId": "price",
            "width": 150,
            "hide": False,
            "pinned": "left",
            "sort": None,
            "sortIndex": None,
            "aggFunc": None,
            "rowGroup": False,
            "rowGroupIndex": None,
            "pivot": False,
            "pivotIndex": None,
            "flex": None,
        },
        {
            "colId": "model",
            "width": 150,
            "hide": False,
            "pinned": None,
            "sort": None,
            "sortIndex": None,
            "aggFunc": None,
            "rowGroup": False,
            "rowGroupIndex": None,
            "pivot": False,
            "pivotIndex": None,
            "flex": None,
        },
    ]

    alt_colState = [
        {
            "colId": "price",
            "width": 198,
            "hide": False,
            "pinned": None,
            "sort": "asc",
            "sortIndex": None,
            "aggFunc": None,
            "rowGroup": False,
            "rowGroupIndex": None,
            "pivot": False,
            "pivotIndex": None,
            "flex": None,
        },
        {
            "colId": "model",
            "width": 150,
            "hide": False,
            "pinned": None,
            "sort": None,
            "sortIndex": None,
            "aggFunc": None,
            "rowGroup": False,
            "rowGroupIndex": None,
            "pivot": False,
            "pivotIndex": None,
            "flex": None,
        },
        {
            "colId": "make",
            "width": 150,
            "hide": False,
            "pinned": None,
            "sort": None,
            "sortIndex": None,
            "aggFunc": None,
            "rowGroup": False,
            "rowGroupIndex": None,
            "pivot": False,
            "pivotIndex": None,
            "flex": None,
        },
    ]

    app.layout = html.Div(
        [
            html.Div(
                [
                    html.Button(
                        "Reset Column State", id="reset-column-state-button", n_clicks=0
                    ),
                    html.Button(
                        "Get Column State", id="get-column-state-button", n_clicks=0
                    ),
                    html.Button(
                        "Load State", id="load-column-state-button", n_clicks=0
                    ),
                    html.Button(
                        "Change Column Defs", id="load-column-defs", n_clicks=0
                    ),
                    html.Button(
                        "Load Both Column Defs and State",
                        id="load-column-state-defs-button",
                        n_clicks=0,
                    ),
                ],
            ),
            dag.AgGrid(
                id="grid",
                # columnSize="autoSize",
                columnDefs=columnDefs,
                defaultColDef=defaultColDef,
                rowData=rowData,
                columnState=colState,
            ),
            html.Div(id="reset-column-state-grid-pre"),
        ]
    )

    @app.callback(
        Output("grid", "resetColumnState"),
        Output("grid", "updateColumnState"),
        Input("reset-column-state-button", "n_clicks"),
        Input("get-column-state-button", "n_clicks"),
    )
    def reset_column_state(n_reset, n_state):
        if ctx.triggered_id == "reset-column-state-button":
            return True, False
        elif ctx.triggered_id == "get-column-state-button":
            return False, True
        return no_update

    @app.callback(
        Output("grid", "columnState"),
        Input("load-column-state-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadState(n):
        if n:
            return colState
        return no_update

    @app.callback(
        Output("reset-column-state-grid-pre", "children"),
        Input("grid", "columnState"),
    )
    def display_column_state(col_state):
        return json.dumps(col_state)

    @app.callback(
        Output("grid", "columnDefs"),
        Input("load-column-defs", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadState(n):
        if n:
            return alt_columnDefs
        return no_update

    @app.callback(
        Output("grid", "columnState", allow_duplicate=True),
        Output("grid", "columnDefs", allow_duplicate=True),
        Input("load-column-state-defs-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def loadState(n):
        if n:
            return colState, columnDefs
        return no_update, no_update

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")

    grid.wait_for_pinned_cols(2)
    grid.wait_for_viewport_cols(1)

    dash_duo.find_element("#get-column-state-button").click()
    time.sleep(0.5)  # pausing to emulate separation because user inputs

    until(
        lambda: json.dumps(colState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )

    grid.resize_col(1, 50)

    dash_duo.find_element("#get-column-state-button").click()
    testState = colState.copy()
    testState[1]["width"] = 198
    until(
        lambda: json.dumps(testState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )

    dash_duo.find_element("#load-column-defs").click()
    until(
        lambda: json.dumps(alt_colState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )
    grid.wait_for_all_header_texts(["Price", "Model", "Make"])
    grid.wait_for_cell_text(0, 0, "32000")

    time.sleep(0.5)  # pausing to emulate separation because user inputs

    dash_duo.find_element("#load-column-state-defs-button").click()
    until(
        lambda: json.dumps(colState)
        in dash_duo.find_element("#reset-column-state-grid-pre").text,
        timeout=3,
    )
    grid.wait_for_all_header_texts(["Make", "Price", "Model"])
